/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.testbase.generator;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.testbase.TestBaseContext;
import ru.ispras.testbase.TestBaseQuery;
import ru.ispras.testbase.TestData;
import ru.ispras.testbase.knowledge.iterator.Iterator;
import ru.ispras.testbase.knowledge.iterator.SingleValueIterator;

import java.util.Set;

/**
 * The {@link DataGeneratorBase} class provides a basic implementation of a data generator.
 * 
 * @author <a href="mailto:andrewt@ispras.ru">Andrei Tatarnikov</a>
 */
public abstract class DataGeneratorBase implements DataGenerator {
  private final String name;
  private final Set<String> instructions;

  /**
   * Constructs a data generator with the specified name applicable to the specified
   * instructions.
   * 
   * @param name Data generator name.
   * @param instructions Names of instructions the generator is applicable to. If it equals
   *        {@code null}, the generator is applicable to all instructions.
   * 
   * @throws IllegalArgumentException if {@code name} is {@code null}.
   */
  public DataGeneratorBase(final String name, final Set<String> instructions) {
    InvariantChecks.checkNotNull(name);

    this.name = name;
    this.instructions = instructions;
  }

  /**
   * Constructs a data generator with the specified name applicable to all instructions.
   * 
   * @param name Data generator name.
   * 
   * @throws IllegalArgumentException if {@code name} is {@code null}.
   */
  public DataGeneratorBase(final String name) {
    this(name, null);
  }

  /**
   * Returns data generator name.
   * 
   * @return Name identifying the data generator.
   */
  public final String getName() {
    return name;
  }

  /**
   * {@inheritDoc}
   * 
   * <p>The generator is considered suitable if two conditions are satisfied:
   * <ol>
   * <li>The generator name matches the test case name specified in the query.</li>
   * <li>The instruction name specified in the query is in the set of supported instructions.
   *     Skipped if the set is {@code null}.</li>
   * </ol>
   */
  @Override
  public boolean isSuitable(final TestBaseQuery query) {
    InvariantChecks.checkNotNull(query);

    if (!name.equals(query.getContext().get(TestBaseContext.TESTCASE))) {
      return false;
    }

    if (null != instructions
        && !instructions.contains(query.getContext().get(TestBaseContext.INSTRUCTION))) {
      return false;
    }

    return true;
  }

  @Override
  public final Iterator<TestData> generate(final TestBaseQuery query) {
    InvariantChecks.checkNotNull(query);
    final DataGeneratorQuery internalQuery = new DataGeneratorQuery(query);

    generate(internalQuery);
    return new SingleValueIterator<>(internalQuery.buildTestData());
  }

  /**
   * Generates data for the specified data generator query. Generated values
   * are assigned to the query variables.
   * 
   * @param query Data generator query to be processed.
   */
  protected abstract void generate(final DataGeneratorQuery query);
}
