/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.testbase.generator;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.util.InvariantChecks;

import java.math.BigInteger;

/**
 * The {@link DataGeneratorVariable} class describes a variable passed to a data generator
 * in a query. Generated data must be assigned to unknown variables.
 * 
 * @author <a href="mailto:andrewt@ispras.ru">Andrei Tatarnikov</a>
 */
public final class DataGeneratorVariable {
  private final String name;
  private final Node node;
  private NodeValue value;

  /**
   * Constructs a variable with the specified name described with the specified
   * {@link Node} object.
   *  
   * @param name Variable name.
   * @param node Node describing the variable.
   * 
   * @throws IllegalArgumentException if any of the parameters is {@code null};
   *         if the node is not a value or a variable; if the variable data type
   *         is not a bit vector.
   */
  public DataGeneratorVariable(final String name, final Node node) {
    InvariantChecks.checkNotNull(name);
    InvariantChecks.checkNotNull(node);

    checkNodeKind(name, node);
    checkDataType(name, node);

    this.name = name;
    this.node = node;
    this.value = null;
  }

  private static void checkNodeKind(final String name, final Node node) {
    if (node.getKind() != Node.Kind.VALUE && node.getKind() != Node.Kind.VARIABLE) {
      throw new IllegalArgumentException(String.format(
          "%s has unsupported node kind %s. Variable or value is expected.", name, node.getKind()));
    }
  }

  private static void checkDataType(final String name, final Node node) {
    if (!node.isType(DataTypeId.BIT_VECTOR)) {
      throw new IllegalArgumentException(String.format(
          "%s has unsupported data type %s. Bit vector is expected.", name, node.getDataType()));
    }
  }

  private void checkDataType(final String name, final DataType type) {
    if (!getType().equals(type)) {
      throw new IllegalArgumentException(String.format(
          "%s is illegal for %s. %s is expected.", type, name, getType()));
    }
  }

  private void checkNotValueSet() {
    if (isValueSet()) {
      throw new IllegalStateException(String.format(
          "%s is already assigned a value.", name));
    }
  }

  /**
   * Returns the variable name.
   * 
   * @return Variable name.
   */
  public String getName() {
    return name;
  }

  /**
   * Returns the variable data type.
   * 
   * @return Variable data type.
   */
  public DataType getType() {
    return node.getDataType();
  }

  /**
   * Returns variable bit size.
   * 
   * @return Variable bit size.
   */
  public int getSize() {
    return getType().getSize();
  }

  /**
   * Checks whether the variable is an unknown variable that must have its value assigned.
   * 
   * @return {@code true} if this is an unknown variable or {@code false} otherwise.
   */
  public boolean isUnknown() {
    return Utils.isUnknownVariable(node);
  }

  /**
   * Checks whether the variable has been assigned a value.
   * 
   * @return {@code true} if the variable has been assigned a value or {@code false} otherwise.
   */
  public boolean isValueSet() {
    return isUnknown() ? null != value : true;
  }

  /**
   * Returns a node describing the variable value.
   * 
   * @return A node describing the variable value.
   * @throws IllegalStateException if the variable has no value.
   */
  public NodeValue getValue() {
    final NodeValue result = isUnknown() ? value : (NodeValue) node;
    if (null == result) {
      throw new IllegalStateException(name + " is not assigned a value.");
    }
    return result;
  }

  /**
   * Returns the variable value as a bit vector.
   * 
   * @return Variable value as a {@link BitVector} object.
   * @throws IllegalStateException if the variable has no value or if it is not
   *         represented by a bit vector.
   */
  public BitVector bitVectorValue() {
    return getValue().getBitVector();
  }

  /**
  * Returns the variable value as an integer value.
  * 
  * @return Variable value as a {@link BigInteger} object.
  * @throws IllegalStateException if the variable has no value or if it cannot be
  *         converted to an integer value.
  */
  public BigInteger integerValue() {
    return bitVectorValue().bigIntegerValue();
  }

  /**
   * Assigns an integer value to the variable. The value is zero-extended
   * to suit the variable size.
   * 
   * @param value value to be assigned.
   * @throws IllegalArgumentException if the argument is {@code null};
   *         if the variable already has a value;
   *         if the value type does not match the variable type.
   */
  public void setValue(final int value) {
    setValue(Data.newBitVector(value, getSize()));
  }

  /**
   * Assigns a long integer value to the variable. The value is zero-extended
   * to suit the variable size.
   * 
   * @param value value to be assigned.
   * @throws IllegalArgumentException if the argument is {@code null};
   *         if the variable already has a value;
   *         if the value type does not match the variable type.
   */
  public void setValue(final long value) {
    setValue(Data.newBitVector(value, getSize()));
  }

  /**
   * Assigns a {@link BigInteger} value to the variable.
   * 
   * @param value {@link BigInteger} object describing the value to be assigned.
   * @throws IllegalArgumentException if the argument is {@code null};
   *         if the variable already has a value;
   *         if the value type does not match the variable type.
   */
  public void setValue(final BigInteger value) {
    setValue(Data.newBitVector(value, getSize()));
  }

  /**
   * Assigns a bit vector value to the variable.
   * 
   * @param value {@link BitVector} object describing the value to be assigned.
   * @throws IllegalArgumentException if the argument is {@code null};
   *         if the variable already has a value;
   *         if the value size does not match the variable size.
   */
  public void setValue(final BitVector value) {
    setValue(NodeValue.newBitVector(value));
  }

  /**
   * Assigns a value described by the specified data object to the variable.
   * 
   * @param value {@link Data} object describing the value to be assigned.
   * @throws IllegalArgumentException if the argument is {@code null};
   *         if the variable already has a value;
   *         if the value type does not match the variable type.
   */
  public void setValue(final Data value) {
    setValue(new NodeValue(value));
  }

  /**
   * Assigns a value described by the specified node to the variable.
   * 
   * @param value {@link NodeValue} object describing the value to be assigned.
   * @throws IllegalArgumentException if the argument is {@code null};
   *         if the variable already has a value;
   *         if the value type does not match the variable type.
   */
  public void setValue(final NodeValue value) {
    InvariantChecks.checkNotNull(value);
    checkNotValueSet();

    checkDataType(name, value.getDataType());
    this.value = value;
  }
}
