/*
 * Copyright 2017 ISP RAS (http://www.ispras.ru)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.testbase.knowledge.iterator;

import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.data.types.bitvector.BitVectorMath;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

/**
 * {@link BoundaryValueIterator} implements a static methods that return iterators for numbers or
 * floating-point value
 * 
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public final class BoundaryValueIterator {
  /** The collection of boundary values sets for different {@link BitVector} sizes. */
  private static Map<Integer, HashSet<BitVector>> boundaryList =
      new HashMap<Integer, HashSet<BitVector>>();

  /**
   * The collection of boundary floating-point values sets for different {@link BitVector} sizes.
   */
  private static Map<Integer, HashSet<BitVector>> boundaryFPList =
      new HashMap<Integer, HashSet<BitVector>>();

  /**
   * Creates a boundary values iterator.
   *
   * @param size the size of {@link BitVector}.
   * @return the boundary values iterator.
   */
  public static CollectionIterator<BitVector> getBoundaryIterator(final int size) {
    HashSet<BitVector> returnedSet;
    if (!boundaryList.containsKey(size)) {
      boundaryList.put(size, BoundaryValueIterator.getBoundaryValues(size));
    }
    returnedSet = boundaryList.get(size);
    return new CollectionIterator<BitVector>(returnedSet);
  }

  /**
   * Creates a boundary floating-point values iterator.
   *
   * @param size the size of {@link BitVector}.
   * @return the boundary floating-point values iterator.
   */
  public static CollectionIterator<BitVector> getFpBoundaryIterator(final int size) {
    HashSet<BitVector> returnedSet;
    if (!boundaryFPList.containsKey(size)) {
      boundaryFPList.put(size, BoundaryValueIterator.getFpBoundaryValues(size));
    }
    returnedSet = boundaryFPList.get(size);
    return new CollectionIterator<BitVector>(returnedSet);
  }

  /*
   * Returns a {@link HashSet} that contains a list of boundary values for floating-point.
   *
   * @param size the number of bits in {@link BitVector}.
   *
   * @return the {@link HashSet} of floating-point boundary values.
   */
  private static HashSet<BitVector> getFpBoundaryValues(final int size) {
    HashSet<BitVector> newSet = new HashSet<BitVector>();
    // TODO:
    return newSet;
  }

  /*
   * Returns a {@link HashSet} that contains a list of boundary values.
   *
   * @param size the number of bits in {@link BitVector}.
   *
   * @return the {@link HashSet} of boundary values.
   */
  private static HashSet<BitVector> getBoundaryValues(final int size) {
    HashSet<BitVector> newSet = new HashSet<BitVector>();

    // 0x0
    newSet.add(BitVector.valueOf(0, size));
    // 0x1
    newSet.add(BitVector.valueOf(1, size));
    // 0x10..0
    BitVector temp = BitVector.valueOf(0, size);
    temp.setBit((size - 1), true);
    newSet.add(temp.copy());
    // 0xf..f
    temp.setAll();
    newSet.add(temp.copy());
    // 0x7f..f
    temp.setBit((size - 1), false);
    newSet.add(temp.copy());
    // 0x10..01
    temp = BitVector.valueOf(1, size);
    temp.setBit((size - 1), true);
    newSet.add(temp.copy());

    BitVector temp2 = BitVector.valueOf(1, size);
    int step = 2; // 2 bits
    while (size / step > 0) {
      temp = temp2;
      temp2 = BitVectorMath.shl(temp2, step / 2);
      temp2 = BitVectorMath.add(temp, temp2);
      newSet.add(temp2);
      step = step * 2;
    }

    return newSet;
  }
}
